/*
	I was helped by looking at the source code of Trevor Blackwell's
	AVR-based unicycle at http://www.tlb.org/eunicycle.html.
*/

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/sleep.h>

#include <stdlib.h>
#include <string.h>

#include "adc.h"
#include "camera.h"
#include "defs.h"
#include "fieldsat.h"
#include "gps.h"
#include "mmc.h"
#include "serial.h"
#include "servo.h"

static void init(void);

static void notify_startup(void);

static void notify_happy(void);

static void notify_mmc_failed(void);

static void get_sample_s1(struct fieldsat_sample_s1 *sample);

static unsigned short get_timer(void);

static unsigned long get_timestamp_high(void);

static unsigned long get_timestamp(void);

static unsigned char seek_empty(unsigned char mmc);

static unsigned int seconds_elapsed(void);

static unsigned int seconds_elapsed_since(unsigned long t);

static unsigned char is_quiescent(void);

static unsigned char should_deploy(void);

static unsigned char should_take_picture(void);

static void deploy(void);

static void write_run_start(void);

static void write_sample_s1(const struct fieldsat_sample_s1 *sample);

static void write_sample_gp(const char *sentence);

static void write_sample_ph(unsigned short num);

static void write_status();

static volatile unsigned short msleep_compare_count;
static unsigned long timestamp_high = 0;

static unsigned char mmc_is_initialized[2] = { 0, 0 };
static unsigned char mmc_write_has_failed[2] = { 0, 0 };

static unsigned char write_buffer[MMC_SECTOR_SIZE];
static unsigned long write_pos = 0;
static unsigned long sector_num = 0;

enum deployment_status {
	DEPLOYMENT_CLOSED,
	DEPLOYMENT_OPENING,
	DEPLOYMENT_OPEN
};

enum deployment_status deployment_status = DEPLOYMENT_CLOSED;

static unsigned char num_pictures;

/* The exponential weighted average of acceleration, used for detecting
   quiescence. */
static float avg_acc[3] = { 512.0, 512.0, 512.0 };

/* The time when the most recent period of quiescence started. */
static unsigned long quiescence_start;
/* The time when deployment started. */
static unsigned long deployment_start;
/* The time when deployment finished or a picture was last taken. */
static unsigned long camera_timer_start;

ISR(__vector_default)
{
	asm volatile ("break");
}

/* This interrupt handler is registered and called during msleep. It decrements
   msleep_compare_count, which is monitored by msleep. */
ISR(TIMER1_COMPA_vect)
{
	if (msleep_compare_count > 0)
		msleep_compare_count--;
}

/* This interrupt handler is called when the 16-bit timer overflows. The 16-bit
   variable timestamp_high is used to increase the range of the timer to 32
   bits. */
ISR(TIMER1_OVF_vect)
{
	timestamp_high++;
}

int main(void)
{
	struct fieldsat_sample_s1 s1;
	unsigned long timestamp, next_timestamp, diff;

	init();

	notify_startup();

#ifdef USE_PARANOID_WRITE
	if (mmc_init(MMC0_CS) && seek_empty(MMC0_CS)) {
#else
	if (mmc_init(MMC0_CS)) {
#endif
		mmc_is_initialized[0] = 1;
		notify_happy();
	} else {
		notify_mmc_failed();
	}

#ifdef USE_PARANOID_WRITE
	if (mmc_init(MMC1_CS) && seek_empty(MMC1_CS)) {
#else
	if (mmc_init(MMC1_CS)) {
#endif
		mmc_is_initialized[1] = 1;
		notify_happy();
	} else {
		notify_mmc_failed();
	}

	write_run_start();

	quiescence_start = get_timestamp_high();

	next_timestamp = get_timestamp();
	for (;;) {
		const char *nmea_sentence;

		get_sample_s1(&s1);
		write_sample_s1(&s1);

		avg_acc[0] += AVG_ACC_RESPONSE * (s1.acc[0] - avg_acc[0]);
		avg_acc[1] += AVG_ACC_RESPONSE * (s1.acc[1] - avg_acc[1]);
		avg_acc[2] += AVG_ACC_RESPONSE * (s1.acc[2] - avg_acc[2]);

		if (!(abs(s1.acc[0] - avg_acc[0]) < QUIESCENCE_THRESHOLD
			&& abs(s1.acc[1] - avg_acc[1]) < QUIESCENCE_THRESHOLD
			&& abs(s1.acc[2] - avg_acc[2]) < QUIESCENCE_THRESHOLD)) {
			quiescence_start = get_timestamp_high();
		}

		if (is_quiescent())
			LED_PORT &= ~_BV(LED);
		else
			LED_PORT |= _BV(LED);

#ifdef USE_GPS
		nmea_sentence = gps_read_sentence();
		if (nmea_sentence != NULL)
			write_sample_gp(nmea_sentence);
#endif

		if (should_deploy()) {
			deployment_status = DEPLOYMENT_OPENING;
			deployment_start = get_timestamp_high();
			deploy();
			write_status();
		}
		if (deployment_status == DEPLOYMENT_OPENING) {
			if (servo_is_enabled()
				&& seconds_elapsed_since(deployment_start) > DEPLOYMENT_DURATION) {
				/* The switch has been pulled, so turn off the
				   servo. */
				servo_disable();
				/* Turn on the unlatching resistor. */
				LATCH_PORT |= _BV(LATCH);
			}
			if (!servo_is_enabled()
				&& seconds_elapsed_since(deployment_start) > DEPLOYMENT_DURATION + UNLATCH_DURATION) {
				deployment_status = DEPLOYMENT_OPEN;
				/* Turn off the unlatching resistor. */
				LATCH_PORT &= ~_BV(LATCH);
				camera_timer_start = get_timestamp_high();
				write_status();
			}
		}

		if (should_take_picture()) {
			num_pictures++;
			write_sample_ph(num_pictures);
			camera_take_picture();
			camera_timer_start = get_timestamp_high();
		}

		#ifdef USE_SAMPLE_RATE_THROTTLE
		/* This is the sample rate throttle. next_timestamp advances in
		   units of SAMPLE_PERIOD, converted from milliseconds to clock
		   cycles. We sleep after each sample long enough to keep the
		   sample rate near SAMPLE_FREQUENCY. */
		/* We want this cycle to end whenever the last cycle ended plus
		   the sample period. */
		next_timestamp += F_CPU / 1000 * SAMPLE_PERIOD;
		timestamp = get_timestamp();
		/* Find out how long we have to wait to get there. */
		diff = next_timestamp - timestamp;
		/* If diff is positive, we need to sleep. If not, we're past
		   where we're supposed to be, so catch up the target
		   timestamp. */
		if ((signed long) diff > 0)
			msleep(diff * 1000 / F_CPU);
		else
			/* If we're already past where we're supposed to be,
			   catch up the target timestamp. */
			next_timestamp = timestamp;
		#endif
	}

	/* Halt the CPU so Avrora knows to quit. */
	asm volatile ("break");

	return 0;
}

static void init(void)
{
	/* Enable interrupts. */
	sei();

	/* Turn on the 16-bit clock with no prescaler. See the ATmega32 data
	   sheet, page 110. */
	TCCR1B = _BV(CS10);
	/* Enable an interrupt when the timer overflows. */
	TIMSK |= _BV(TOIE1);

	/* Set the LED pin to output. */
	LED_DDR |= _BV(LED);
	/* Turn off the LED by setting its bit. */
	LED_PORT |= _BV(LED);

	/* Set the latch pin to output. */
	LATCH_DDR |= _BV(LATCH);
	LATCH_PORT &= ~_BV(LATCH);

	serial_init();

#ifdef USE_GPS
	gps_init();
#endif

	adc_init();

	servo_init();
	servo_disable();

	camera_init();
	num_pictures = 0;
}

/* Display some sort of notification message. */
static void notify_startup(void)
{
	notify_happy();
}

/* Give a happy notification, the opposite of an error. */
static void notify_happy(void)
{
	unsigned char i;

	for (i = 0; i < 3; i++) {
		flash(50);
		msleep(50);
	}
}

static void notify_mmc_failed(void)
{
	flash(300);
	msleep(300);
	flash(300);
	msleep(300);
}

/* Sleep for ms milliseconds. This function uses the timer 1 output compare
   register and interrupt. That register and interrupt are reserved for this
   function. */
void msleep(unsigned short ms)
{
	unsigned long num_cycles;
	unsigned short start_time;
	unsigned char sreg;

	start_time = get_timer();

	num_cycles = F_CPU / 1000 * ms;

	/* Set the output compare register 1 A to the low 16 bits of num_cycles
	   after the current time. See the note on accessing 16-bit registers in
	   get_timer. */
	sreg = SREG;
	cli();
	OCR1A = start_time + num_cycles;
	SREG = sreg;

	/* Loop for the high 16 bits of the number of cycles, plus one. */
	msleep_compare_count = (num_cycles >> 16) + 1;

	/* Enable an interrupt on timer 1 output compare match. */
	TIMSK |= _BV(OCIE1A);

	set_sleep_mode(SLEEP_MODE_IDLE);
	/* msleep_compare_count is modified by the interrupt handler. */
	while (msleep_compare_count > 0)
		sleep_mode();

	/* Disable the interrupt. */
	TIMSK &= ~_BV(OCIE1A);
}

/* Light the LED for ms milliseconds. */
void flash(unsigned short ms)
{
	LED_PORT &= ~_BV(LED);
	msleep(ms);
	LED_PORT |= _BV(LED);
}

static void get_sample_s1(struct fieldsat_sample_s1 *sample)
{
	memset(sample, 0, sizeof(*sample));
	sample->timestamp = get_timestamp();
	/* Read the first sample twice. This gets rid of weird glitches that
	   appear otherwise. I don't know what causes this. */
	sample->ang_vel[0] = adc_read_channel(CHANNEL_X_ANG_VEL);
	sample->ang_vel[0] = adc_read_channel(CHANNEL_X_ANG_VEL);
	sample->ang_vel[1] = adc_read_channel(CHANNEL_Y_ANG_VEL);
	sample->ang_vel[2] = adc_read_channel(CHANNEL_Z_ANG_VEL);
	sample->acc[0] = adc_read_channel(CHANNEL_X_ACC);
	sample->acc[1] = adc_read_channel(CHANNEL_Y_ACC);
	sample->acc[2] = adc_read_channel(CHANNEL_Z_ACC);
	/*
	sample->pressure = adc_read_channel(CHANNEL_PRESSURE);
	sample->temp_ext = adc_read_channel(CHANNEL_TEMP_EXT);
	*/
}

/* Return the current value of the 16-bit timer. */
static unsigned short get_timer(void)
{
	unsigned char sreg;
	unsigned short timer;

	/* TCNT1 is a 16-bit register. It has to be read in a special way to
	   prevent the read being interrupted by an interrupt. See the ATmega32
	   data sheet, page 89. */
	/* Save the status register (including the interrupt flag). */
	sreg = SREG;
	/* Disable interrupts. */
	cli();
	/* Read the 16-bit value. */
	timer = TCNT1;
	/* Restore the status register. */
	SREG = sreg;

	return timer;
}

/* Return the value of the 32-bit timer timestamp_high. */
static unsigned long get_timestamp_high(void)
{
	unsigned char sreg;
	unsigned long timestamp;

	sreg = SREG;
	cli();
	timestamp = timestamp_high;
	SREG = sreg;

	return timestamp;
}

/* Return a 32-bit timestamp. The high 16 bits are the value of timestamp_high
   and the low 16 bits are the current value of the timer. */
static unsigned long get_timestamp(void)
{
	unsigned char sreg;
	unsigned long timestamp;

	sreg = SREG;
	/* Disable interrupts to keep timestamp_high from being modified while
	   we're reading it. Otherwise, there is a race condition. */
	cli();
	timestamp = get_timer();
	if (TIFR & _BV(TOV1)) {
		/* If the TOV1 (Timer 1 overflow 1) bit of the TIFR (timer
		   interrupt flag register) is set, we missed an overflow.
		   Pretend we're at the end of this cycle. This guarantees that
		   the timestamp sequence will be at least non-decreasing. The
		   overflow interrupt will run as soon as we re-enable
		   interrupts. */
		timestamp = 0xFFFF;
	}
	timestamp |= timestamp_high << 16;
	SREG = sreg;

	return timestamp;
}

/* Look for the first all-zero sector, starting at sector_num. */
static unsigned char seek_empty(unsigned char mmc)
{
	unsigned int i;

	for (;;) {
		if (!mmc_read_sector(mmc, write_buffer, sector_num))
			return 0;
		for (i = 0; i < MMC_SECTOR_SIZE; i++) {
			if (write_buffer[i] != 0x00)
				break;
		}
		if (i >= MMC_SECTOR_SIZE)
			return 1;
		sector_num++;
	}
}

static unsigned int seconds_elapsed(void)
{
	return seconds_elapsed_since(0);
}

/* Return the number of seconds elapsed since t, which is a timestamp as
   returned by get_timestamp_high. */
static unsigned int seconds_elapsed_since(unsigned long t)
{
	unsigned int elapsed;
	unsigned char sreg;

	sreg = SREG;
	cli();
	elapsed = (timestamp_high - t) / (F_CPU >> 16);
	SREG = sreg;

	return elapsed;
}

static unsigned char is_quiescent(void)
{
	return seconds_elapsed_since(quiescence_start) > 1;
}

static unsigned char should_deploy(void)
{
	return deployment_status == DEPLOYMENT_CLOSED
		&& seconds_elapsed() > DEPLOYMENT_TIME
		&& seconds_elapsed_since(quiescence_start) > QUIESCENCE_TIMER;
}

static unsigned char should_take_picture(void)
{
	if (deployment_status != DEPLOYMENT_OPEN || num_pictures >= MAX_NUM_PICTURES)
		return 0;
	if (num_pictures == 0)
		return 1;
	return seconds_elapsed_since(camera_timer_start) > CAMERA_TIME;
}

static void deploy(void)
{
	servo_enable();
	servo_set_position(SERVO_NONE);
}

static void write(const unsigned char *buffer, unsigned int size);
static void write_char(char c);
static void write_string(const char *s);
static void write_hex_long(unsigned long l);
static void write_hex_short(unsigned short h);

static void write_timestamp(unsigned long timestamp)
{
	write_string("TS=");
	write_hex_long(timestamp);
}

static void write_run_start(void)
{
	write_string("# Metro State Field/GPSSat data file format 2\n");
	write_string("# http://www.bamsoftware.com/wiki/DemoSatV/\n");
	write_status();
}

static void write_sample_s1(const struct fieldsat_sample_s1 *sample)
{
	write_timestamp(sample->timestamp);
	write_char(' ');
	write_string("RX=");
	write_hex_short(sample->ang_vel[0]);
	write_char(' ');
	write_string("RY=");
	write_hex_short(sample->ang_vel[1]);
	write_char(' ');
	write_string("RZ=");
	write_hex_short(sample->ang_vel[2]);
	write_char(' ');
	write_string("AX=");
	write_hex_short(sample->acc[0]);
	write_char(' ');
	write_string("AY=");
	write_hex_short(sample->acc[1]);
	write_char(' ');
	write_string("AZ=");
	write_hex_short(sample->acc[2]);
	write_char('\n');
}

static void write_sample_gp(const char *sentence)
{
	write_timestamp(get_timestamp());
	write_char(' ');
	write_string("GPS=");
	write_string(sentence);
	write_char('\n');
}

static void write_sample_ph(unsigned short num)
{
	write_timestamp(get_timestamp());
	write_char(' ');
	write_string("PHOTO=");
	write_hex_short(num);
	write_char('\n');
}

static void write_status()
{
	write_timestamp(get_timestamp());
	write_char(' ');
	write_string("DEPLOY=");
	switch (deployment_status) {
		case DEPLOYMENT_CLOSED:
			write_string("CLOSED");
			break;
		case DEPLOYMENT_OPENING:
			write_string("OPENING");
			break;
		case DEPLOYMENT_OPEN:
			write_string("OPEN");
			break;
		default:
			write_string("UNKNOWN");
			break;
	}
	write_char(' ');
	write_string("NUMPHOTOS=");
	write_hex_short(num_pictures);
	write_char('\n');
}

static void write(const unsigned char *buffer, unsigned int size)
{
	unsigned int offset;

	offset = 0;
	while (size - offset >= sizeof(write_buffer) - write_pos) {
		memmove(write_buffer + write_pos, buffer + offset, sizeof(write_buffer) - write_pos);
		offset += sizeof(write_buffer) - write_pos;
		if (mmc_is_initialized[0]) {
			if (!mmc_write_sector(MMC0_CS, write_buffer, sector_num)
				&& !mmc_write_has_failed[0]) {
				notify_mmc_failed();
				mmc_write_has_failed[0] = 1;
			}
		}
		if (mmc_is_initialized[1]) {
			if (!mmc_write_sector(MMC1_CS, write_buffer, sector_num)
				&& !mmc_write_has_failed[1]) {
				notify_mmc_failed();
				mmc_write_has_failed[1] = 1;
			}
		}
		write_pos = 0;
		sector_num++;
		/* Beep the first NUM_CONFIRMATION_FLASHES times that
		   NUM_CONFIRMATION_SECTORS are written. */
		if ((sector_num / NUM_CONFIRMATION_SECTORS) <= NUM_CONFIRMATION_FLASHES
			&& sector_num % NUM_CONFIRMATION_SECTORS == 0)
			flash(20);
	}
	memmove(write_buffer + write_pos, buffer + offset, size - offset);
	write_pos += size - offset;
}

static void write_char(char c)
{
	write((unsigned char *) &c, 1);
}

static void write_string(const char *s)
{
	write((unsigned char *) s, strlen(s));
}

static const char *HEX_DIGITS = "0123456789ABCDEF";

static void write_hex_long(unsigned long l)
{
	unsigned char n = 32;
	char c;

	do {
		n -= 4;
		c = HEX_DIGITS[(l >> n) & 0x0F];
		write_char(c);
	} while (n > 0);
}

static void write_hex_short(unsigned short h)
{
	unsigned char n = 16;
	char c;

	do {
		n -= 4;
		c = HEX_DIGITS[(h >> n) & 0x0F];
		write_char(c);
	} while (n > 0);
}
