/* This file contains functions for reading and writing an MMC card in SPI
   mode. In addition to the sources mentioned below, I consulted the web page at
   http://www.captain.at/electronic-atmega-mmc.php for example code. This code
   is quite different, however. */

#include <avr/io.h>

#include "defs.h"
#include "mmc.h"

/*
#define SPI_DEBUG
*/

static void spi_init(void);

static unsigned char spi_send_recv(unsigned char c);

static void spi_send(unsigned char c);

static unsigned char spi_recv(void);

/* The ATmega32 SPI (Serial Peripheral Interface), used by the MMC driver, is
   discussed starting on page 132 of the ATmega32 data sheet. See also
   application note AVR151: Setup and Use of The SPI. */

/* Initialize the SPI interface. See AVR151, page 10. */
static void spi_init(void)
{
	volatile unsigned char dummy;

	/* Set these pins to output:
	       PB4 (SS)
	       PB5 (MOSI)
	       PB7 (SCK)
	   We must set PB4 to output even if it is not used as a slave select
	   pin. See the ATmega32 data sheet, page 136. */
	DDRB |= _BV(PB4) | _BV(PB5) | _BV(PB7);
	/* Set pin PB6 (MISO) to input. This happens automatically when the SPI
	   interface is enabled in master mode. */
	DDRB &= ~_BV(PB6);

	/* Enable SPI in master mode. */
	SPCR |= _BV(SPE) | _BV(MSTR);
	/* These other bits control SPI parameters. See the ATmega32 data sheet,
	   pages 136-138. */
	/* The DORD bit of SPCR selects the data order. DORD = 0 means send the
	   MSB first, which is what we want for the MMC. */
	/*
	SPCR |= _BV(DORD);
	*/
	/* Bit CPOL is clock polarity and CPHA is clock phase. The defaults (low
	   when idle, sample on leading edge) are correct. */
	/*
	SPCR |= _BV(CPOL);
	SPCR |= _BV(CPHA);
	*/
	/* Bits SPR1 and SPR0 of SPCR select the SPI clock rate. SPR1 = SPR0 = 0
	   divides the processor clock rate by 4. SPR1 = SPR0 = 1 divides the
	   processor clock by 128. */
	/*
	*/
	SPCR |= _BV(SPR1);
	SPCR |= _BV(SPR0);

	/* Clear the SPIF (SPI interrupt flag) bit by reading SPSR and SPDR. */
	dummy = SPSR;
	dummy = SPDR;
}

#ifdef SPI_DEBUG
#include <stdlib.h>
#include "serial.h"
#endif

/* With SPI, sends and receives always happen simultaneously. Send a character
   and return the character received. */
static unsigned char spi_send_recv(unsigned char c)
{
#ifdef SPI_DEBUG
	char buf[3];

	serial_puts("SPI: ");
	utoa(c, buf, 16);
	if (buf[1] == '\0')
		serial_putchar('0');
	serial_puts(buf);
	serial_putchar(' ');
#endif
	SPDR = c;
	loop_until_bit_is_set(SPSR, SPIF);
	c = SPDR;
#ifdef SPI_DEBUG
	utoa(c, buf, 16);
	if (buf[1] == '\0')
		serial_putchar('0');
	serial_puts(buf);
	serial_putchar('\n');
#endif

	return c;
}

static void spi_send(unsigned char c)
{
	spi_send_recv(c);
}

static unsigned char spi_recv(void)
{
	/* Send a dummy byte. Send 0xFF because parts of the MMC protocol
	   require holding the MOSI line high while reading. */
	return spi_send_recv(0xFF);
}

/* The source I consulted for the MMC protocol is the SanDisk MultiMediaCard and
   Reduced-Size MultiMediaCard Product Manual, version 1.3, available at
   http://www.sandisk.com/Assets/File/OEM/Manuals/ProdManRS-MMCv1.3.pdf. The
   discussion of using MMC cards in SPI mode starts on page 66. */

static unsigned char mmc_command(unsigned char command, unsigned long argument);

static unsigned char mmc_spi_mode(void);

static void mmc_select_none();

static void mmc_select(unsigned char mask);

static unsigned char mmc_read_data(unsigned char *buffer, unsigned long size);

enum mmc_command {
	CMD_GO_IDLE_STATE = 0,
	CMD_SEND_OP_COND = 1,
	CMD_SEND_CSD = 9,
	CMD_SEND_CID = 10,
	CMD_STOP_TRANSMISSION = 12,
	CMD_SEND_STATUS = 13,
	CMD_SET_BLOCKLEN = 16,
	CMD_READ_SINGLE_BLOCK = 17,
	CMD_READ_MULTIPLE_BLOCK = 18,
	CMD_WRITE_BLOCK = 24,
	CMD_WRITE_MULTIPLE_BLOCK = 25,
	CMD_PROGRAM_CSD = 27,
	CMD_SET_WRITE_PROT = 28,
	CMD_CLR_WRITE_PROT = 29,
	CMD_SEND_WRITE_PROT = 30,
	CMD_TAG_SECTOR_START = 32,
	CMD_TAG_SECTOR_END = 33,
	CMD_UNTAG_SECTOR = 34,
	CMD_TAG_ERASE_GROUP_START = 35,
	CMD_TAG_ERASE_GROUP_END = 36,
	CMD_UNTAG_ERASE_GROUP = 37,
	CMD_ERASE = 38,
	CMD_LOCK_UNLOCK = 42,
	CMD_READ_OCR = 58,
	CMD_CRC_ON_OFF = 59
};

/* The number of bytes to wait for certain responses before giving up. */
#define MMC_TIMEOUT 10000
/* The number of bytes to wait for a block to be written before giving up. */
#define MMC_WRITE_TIMEOUT 10000

/* Send an MMC command. A dummy checksum is also sent. Return the R1 response,
   or 0xFF on timeout. Other response types are not supported. See the MMC
   manual, pages 52 (Command Format), 78 (R1 Format), and 83 (Command and
   Response). */
static unsigned char mmc_command(unsigned char command, unsigned long argument)
{
	unsigned int i;
	unsigned char c;

	/* Push the MOSI line high before sending the command. */
	spi_send(0xFF);
	spi_send((command & 0x3F) | 0x40);
	spi_send((unsigned char) (argument >> 24));
	spi_send((unsigned char) (argument >> 16));
	spi_send((unsigned char) (argument >> 8));
	spi_send((unsigned char) argument);
	/* Send a dummy checksum. */
	spi_send(0xFF);

	/* Loop until we get a return code. */
	for (i = 0; i < MMC_TIMEOUT; i++) {
		c = spi_recv();
		if (c != (unsigned char) 0xFF)
			break;
	}
	if (i == MMC_TIMEOUT)
		return 0xFF;

	return c;
}

/* Cause the MMC card to enter SPI mode. The card's slave select pin must be low
   before calling this function. This function is like mmc_command with constant
   arguments. */
static unsigned char mmc_spi_mode(void)
{
	unsigned int i;
	unsigned char c;

	/* Send CMD0, GO_IDLE_STATE. This command is constant with a precomputed
	   CRC. Sending this command while the slave select pin is low puts the
	   card into SPI mode. See the MMC manual, page 69 (Bus Transfer
	   Protection). */
	/* Push the MOSI line high before sending the command. */
	spi_send(0xFF);
	spi_send((CMD_GO_IDLE_STATE & 0x3F) | 0x40);
	spi_send(0x00);
	spi_send(0x00);
	spi_send(0x00);
	spi_send(0x00);
	spi_send(0x95);

	/* Loop until we get a return code. */
	for (i = 0; i < MMC_TIMEOUT; i++) {
		c = spi_recv();
		if (c != (unsigned char) 0xFF)
			break;
	}
	if (i == MMC_TIMEOUT)
		return 0xFF;

	return c;
}

static void mmc_select_none(void)
{
#ifdef SPI_DEBUG
	serial_puts("MMC: (none)\n");
#endif
	MMC_PORT &= ~(_BV(MMC0_CS) | _BV(MMC1_CS));
}

static void mmc_select(unsigned char mmc)
{
#ifdef SPI_DEBUG
	char buf[3];

	serial_puts("MMC: ");
	utoa(mmc, buf, 10);
	serial_puts(buf);
	serial_putchar('\n');
#endif
	/* SPI slave selects are active-low. */
	if (mmc == MMC0_CS) {
		MMC_PORT &= ~_BV(MMC0_CS);
		MMC_PORT |= _BV(MMC1_CS);
	} else if (mmc == MMC1_CS) {
		MMC_PORT &= ~_BV(MMC1_CS);
		MMC_PORT |= _BV(MMC0_CS);
	}
}

/* Initialize an MMC card. This puts the card into SPI mode. */
unsigned char mmc_init(unsigned char mmc)
{
	unsigned char i;
	unsigned char result;

	spi_init();

	/* Set the slave select pins to output. */
	MMC_DDR |= _BV(MMC0_CS) | _BV(MMC1_CS);

	/* Set the slave select pins high to disable all the cards. */
	mmc_select_none();

	/* Send an initialization sequence of 80 '1's. We need to send at least
	   74 to synchronize after power-up. See the MMC manual, page 20
	   (Power Up). This is safe to do even if some cards are already
	   initialized. */
	for (i = 0; i < 10; i++)
		spi_send(0xFF);

	/* Set the slave select pin low to select the card. */
	mmc_select(mmc);

	/* Enter SPI mode. */
	result = mmc_spi_mode();
	/* The idle bit (bit 0) should be the only one set. */
	if (result != 0x01)
		return 0;

	/* Loop until the idle bit (bit 0) is cleared. */
	while ((result = mmc_command(CMD_SEND_OP_COND, 0UL)) == 0x01)
		;
	if (result != 0x00)
		return 0;

	/* The block size is by default the sector size of 512. */

	return 1;
}

/* Read a data token of size bytes. Discard the CRC. Return zero for failure,
   nonzero for success. See the MMC manual, pages 80 and 81 (Data Tokens and
   Data Error Token). */
static unsigned char mmc_read_data(unsigned char *buffer, unsigned long size)
{
	volatile unsigned char dummy;
	unsigned int i;
	unsigned char c;

	for (i = 0; i < MMC_TIMEOUT; i++) {
		c = spi_recv();
		if (c != (unsigned char) 0xFF)
			break;
	}
	/* 0xFE is the start of a data token. Anything else is an error. */
	if (i == MMC_TIMEOUT || c != (unsigned char) 0xFE)
		return 0;

	for ( ; size > 0; size--)
		*(buffer++) = spi_recv();

	/* Discard the CRC. */
	dummy = spi_recv();
	dummy = spi_recv();

	return 1;
}

/* Get the card identification register (CID). See the MMC manual, page 24 (Card
   Identification Register). */
unsigned char mmc_get_cid(unsigned char mmc, unsigned char cid[16])
{
	unsigned char result;

	mmc_select(mmc);
	result = mmc_command(CMD_SEND_CID, 0UL);
	if (result != 0x00)
		return 0;

	return mmc_read_data(cid, 16);
}

/* Get the status register. See the MMC manual, page 79 (R2 Format). */
unsigned short mmc_get_status(unsigned char mmc)
{
	unsigned short status;

	mmc_select(mmc);
	status = mmc_command(CMD_SEND_STATUS, 0UL);

	return (status << 8) | spi_recv();
}

/* Read a sector at address. Return the block in buffer. See the MMC manual,
   pages 69 (Data Read) and 84-85 (Data Read). */
unsigned char mmc_read_sector(unsigned char mmc,
	unsigned char buffer[MMC_SECTOR_SIZE], unsigned long address)
{
	unsigned char result;

	mmc_select(mmc);
	result = mmc_command(CMD_READ_SINGLE_BLOCK, address * MMC_SECTOR_SIZE);
	if (result != 0x00)
		return 0;

	return mmc_read_data(buffer, MMC_SECTOR_SIZE);
}

/* Write a block at address. See the MMC manual, pages 70-71 (Data Write) and
   85-86 (Data Write). */
unsigned char mmc_write_sector(unsigned char mmc,
	const unsigned char buffer[MMC_SECTOR_SIZE], unsigned long address)
{
	unsigned char result;
	unsigned char response;
	unsigned int i;

	mmc_select(mmc);
	result = mmc_command(CMD_WRITE_BLOCK, address * MMC_SECTOR_SIZE);
	if (result != 0x00)
		return 0;

	/* Send the start block byte. See the MMC manual, page 81 (Data
	   Tokens). */
	spi_send(0xFE);
	for (i = 0; i < MMC_SECTOR_SIZE; i++)
		spi_send(buffer[i]);
	/* Send a dummy CRC. */
	spi_send(0xFF);
	spi_send(0xFF);

	/* Get the data response token. See the MMC manual, page 80 (Data
	   Response). */
	response = (spi_recv() & 0x1F) >> 1;
	/* A status of 0x02 means "data accepted." Anything else is an error. */
	if (response != 0x02)
		return 0;

	/* The card will send a continuous string of zeros while it is busy
	   writing. */
	for (i = 0; i < MMC_WRITE_TIMEOUT; i++) {
		if (spi_recv() != (unsigned char) 0x00)
			break;
	}
	if (i == MMC_WRITE_TIMEOUT)
		return 0;

	return 1;
}
