/* The ATmega32 USART (Universal synchronous and asynchronous receiver and
   transmitter) is discussed starting on page 140 of the ATmega32 data sheet. */

/* In order to receive 8-bit characters on your Linux computer, do something
   like this:
   	stty -F /dev/ttyS0 raw 38400
*/

#include <avr/interrupt.h>
#include <avr/io.h>

#include <stddef.h>
#include <string.h>

#include "defs.h"

#ifdef USE_GPS
/* This Baud rate is used by the GPS receiver. */
#define SERIAL_BAUD_RATE 4800UL
#else
#define SERIAL_BAUD_RATE 38400UL
#endif

/* See the ATmega32 data sheet, page 143 for the source of this formula. */
#define UBRR_BAUD (F_CPU / (16 * SERIAL_BAUD_RATE) - 1)

#if defined (__AVR_ATmega128__)
#define UBRRH UBRR0H
#define UBRRL UBRR0L
#define UCSRA UCSR0A
#define UCSRB UCSR0B
#define UDR UDR0
#endif

#define RECEIVE_BUFFER_SIZE 256

/* This is a circular buffer of received bytes. It is filled in by the
   USART_RXC_vect interrupt handler. */
static char receive_buffer[RECEIVE_BUFFER_SIZE];
static volatile unsigned char receive_head = 0;
static unsigned char receive_tail = 0;

/* This interrupt handler is called whenever there is data in the USART receive
   buffer. It has no effect unless the RXCIE (RX Complete Interrupt Enable) bit
   is set in the UCSRB register. See the ATmega32 data sheet, page 161. */
ISR(USART_RXC_vect)
{
	unsigned char next_head;
	volatile char c;

	c = UDR;
	next_head = (receive_head + 1) % sizeof(receive_buffer);
	/* Ignore this byte if we're out of room. */
	if (next_head == receive_head)
		return;
	receive_buffer[receive_head] = c;
	receive_head = next_head;
}

/* For USART initialization, see the ATmega32 data sheet, page 146. */
void serial_init(void)
{
	/* Set the baud rate. The UBRRH register shares the same location as the
	   UCSRC register. Bit 7 (URSEL) must be zero to write to UBRRH. See the
	   ATmega32 data sheet, page 158. */
	UBRRH = (unsigned char) (UBRR_BAUD >> 8);
	UBRRL = (unsigned char) UBRR_BAUD;
	/* Set the RXCIE (RX Complete Interrupt Enable), RXEN (Receiver Enable),
	   and TXEN (Transmitter Enable) bits on the USART. */
	UCSRB |= _BV(RXCIE) | _BV(RXEN) | _BV(TXEN);
	/* The UCSRC register controls synchronous/asynchronous mode, parity and
	   stop bits, etc. The defaults are fine. */
}

void serial_putchar(char c)
{
	/* Wait until the UDRE (USART Data Register Empty) bit is set. See the
	   ATmega32 data sheet, page 160. */
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = c;
}

void serial_write(const unsigned char *buffer, unsigned int size)
{
	for ( ; size > 0; size--)
		serial_putchar(*buffer++);
}

void serial_puts(const char *s)
{
	for ( ; *s != '\0'; s++)
		serial_putchar(*s);
}

char serial_getchar(void)
{
	char c;

	/* Loop until there's something in the buffer. */
	while (receive_tail == receive_head)
		;

	c = receive_buffer[receive_tail];

	receive_tail = (receive_tail + 1) % sizeof(receive_buffer);

	return c;
}

int serial_read(char *buf, size_t n)
{
	size_t count, len;

	count = 0;
	if (receive_tail > receive_head && count < n) {
		len = sizeof(receive_buffer) - receive_tail;
		if (len > n)
			len = n;
		memcpy(buf + count, receive_buffer + receive_tail, len);
		receive_tail = (receive_tail + len) % sizeof(receive_buffer);
		count += len;
	}
	if (receive_tail < receive_head && count < n) {
		len = receive_head - receive_tail;
		if (len > n)
			len = n;
		memcpy(buf + count, receive_buffer + receive_tail, len);
		receive_tail = (receive_tail + len) % sizeof(receive_buffer);
		count += len;
	}

	return count;
}
