#include <avr/io.h>

#include "defs.h"
#include "servo.h"

/* Initialize the servo. To use it, you still have to enable it with
   servo_enable. */
void servo_init(void)
{
	/* Set the servo control pin to output. */
	SERVO_DDR |= _BV(SERVO);
	/* Turn on timer/counter 0 with a prescaler of 256. This gives a
	   frequency of F_CPU / (256 * 256) = 122.07 Hz when F_CPU = 8000000.
	   The servo needs at least 50 Hz to operate.  See the ATmega32 data
	   sheet, page 82. */
	TCCR0 |= _BV(CS02);
	/* Set the waveform generation mode bits (WGM01 and WGM00) to fast PWM
	   mode. See the ATmega32 data sheet, page 80. */
	TCCR0 |= _BV(WGM01) | _BV(WGM00);
}

void servo_enable(void)
{
	/* Set the output pin (OC0) to clear on compare match and set on TOP.
	   See the ATmega32 data sheet, page 81. */
	TCCR0 &= ~_BV(COM00);
	TCCR0 |= _BV(COM01);
}

/* Disconnect the servo control pin so that the servo will not attempt to hold
   any position. */
void servo_disable(void)
{
	/* Set the output pin (OC0) to normal port operation (no PWM).  See the
	   ATmega32 data sheet, page 81. */
	TCCR0 &= ~(_BV(COM01) | _BV(COM00));
}

unsigned char servo_is_enabled(void)
{
	return (TCCR0 & (_BV(COM01) | _BV(COM00))) != 0;
}

void servo_set_position(enum servo_position pos)
{
	/* The servo works by receiving a pulse 50 times a second, or more
	   frequently. A pulse of 1 ms moves the servo all the way
	   counterclockwise, and a pulse of 2 ms moves the servo all the way
	   clockwise. With a prescaler of 256, the frequency of the PWM channel
	   is F_CPU / (256 * 256) (see the ATmega32 data sheet, page 76). Then
	   the period is 256 * 256 / F_CPU. So with F_CPU = 8000000 (8 MHz), a
	   counterclockwise pulse should be
	   0.001 / (256 * 256 / F_CPU) * 256 = 31.25 (rounds to 31)
	   And a clockwise pulse should be
	   0.002 / (256 * 256 / F_CPU) * 256 = 62.50 (rounds to 63).

	   But in actual practice, these values don't cover the whole range.
	   These values are based on trial and error to find the approximate
	   limits. */
	if (pos == SERVO_FULL)
		OCR0 = 72;
	else if (pos == SERVO_HALF)
		OCR0 = 44;
	else if (pos == SERVO_QUARTER)
		OCR0 = 32;
	else
		OCR0 = 22;
}
